import json

import requests

ITEMS_URL = "https://secure.runescape.com/m=itemdb_oldschool/api"
PRICES_URL = "https://prices.runescape.wiki/api/v1/osrs"
HISCORES_URL = "https://secure.runescape.com/m=hiscore_oldschool/index_lite.json"
WIKI_URL = "https://oldschool.runescape.wiki/api.php"
USER_AGENT = "YEW - https://github.com/bharat-nair/yew"

ETAG = ""
ITEMS = {}

# Fetch item mapping (ID-Item) from prices.runescape.com
# Save mapping in a local JSON file as a cache
# Check ETag value for future requests to reduce network load
response = requests.get(
    f"{PRICES_URL}/mapping",
    headers={"User-Agent": USER_AGENT, "If-None-Match": ETAG},
)
if response.status_code == 200:
    ETAG = response.headers.get("ETag", "")
    items = response.json()

    with open("items.json", "w") as f:
        for item in items:
            ITEMS[int(item["id"])] = item



from dataclasses import dataclass

import requests

from yew import HISCORES_URL, USER_AGENT


@dataclass
class Skill:
    rank: int
    level: int
    xp: int


SKILLS = [
    "Overall",
    "Attack",
    "Defence",
    "Strength",
    "Hitpoints",
    "Ranged",
    "Prayer",
    "Magic",
    "Cooking",
    "Woodcutting",
    "Fletching",
    "Fishing",
    "Firemaking",
    "Crafting",
    "Smithing",
    "Mining",
    "Herblore",
    "Agility",
    "Thieving",
    "Slayer",
    "Farming",
    "Runecraft",
    "Hunter",
    "Construction",
]


@dataclass
class Activity:
    rank: int
    score: int


ACTIVITIES = [
    "League Points",
    "Deadman Points",
    "Bounty Hunter - Hunter",
    "Bounty Hunter - Rogue",
    "Bounty Hunter (Legacy) - Hunter",
    "Bounty Hunter (Legacy) - Rogue",
    "Clue Scrolls (all)",
    "Clue Scrolls (beginner)",
    "Clue Scrolls (easy)",
    "Clue Scrolls (medium)",
    "Clue Scrolls (hard)",
    "Clue Scrolls (elite)",
    "Clue Scrolls (master)",
    "LMS - Rank",
    "PvP Arena - Rank",
    "Soul Wars Zeal",
    "Rifts closed",
    "Abyssal Sire",
    "Alchemical Hydra",
    "Artio",
    "Barrows Chests",
    "Bryophyta",
    "Callisto",
    "Calvar'ion",
    "Cerberus",
    "Chambers of Xeric",
    "Chambers of Xeric: Challenge Mode",
    "Chaos Elemental",
    "Chaos Fanatic",
    "Commander Zilyana",
    "Corporeal Beast",
    "Crazy Archaeologist",
    "Dagannoth Prime",
    "Dagannoth Rex",
    "Dagannoth Supreme",
    "Deranged Archaeologist",
    "Duke Sucellus",
    "General Graardor",
    "Giant Mole",
    "Grotesque Guardians",
    "Hespori",
    "Kalphite Queen",
    "King Black Dragon",
    "Kraken",
    "Kree'Arra",
    "K'ril Tsutsaroth",
    "Mimic",
    "Nex",
    "Nightmare",
    "Phosani's Nightmare",
    "Obor",
    "Phantom Muspah",
    "Sarachnis",
    "Scorpia",
    "Scurrius",
    "Skotizo",
    "Spindel",
    "Tempoross",
    "The Gauntlet",
    "The Corrupted Gauntlet",
    "The Leviathan",
    "The Whisperer",
    "Theatre of Blood",
    "Theatre of Blood: Hard Mode",
    "Thermonuclear Smoke Devil",
    "Tombs of Amascut",
    "Tombs of Amascut: Expert Mode",
    "TzKal-Zuk",
    "TzTok-Jad",
    "Vardorvis",
    "Venenatis",
    "Vet'ion",
    "Vorkath",
    "Wintertodt",
    "Zalcano",
    "Zulrah",
]


class PlayerFetcher:
    @staticmethod
    def fetch_player_by_name(name: str):
        url = f"{HISCORES_URL}?player={name}"

        response = requests.get(url, headers={"User-Agent": USER_AGENT}).json()

        player_skills = {}
        player_activities = {}
        for skill in response.get("skills", [{}]):
            player_skills[skill["name"]] = skill
        for activity in response.get("activities", [{}]):
            player_activities[activity["name"]] = activity

        return {"skills": player_skills, "activities": player_activities}


@dataclass
class Player:

    overall: Skill
    attack: Skill
    defence: Skill
    strength: Skill
    hitpoints: Skill
    ranged: Skill
    prayer: Skill
    magic: Skill
    cooking: Skill
    woodcutting: Skill
    fletching: Skill
    fishing: Skill
    firemaking: Skill
    crafting: Skill
    smithing: Skill
    mining: Skill
    herblore: Skill
    agility: Skill
    thieving: Skill
    slayer: Skill
    farming: Skill
    runecraft: Skill
    hunter: Skill
    construction: Skill

    league_points: Activity
    deadman_points: Activity
    bounty_hunter__hunter: Activity
    bounty_hunter__rogue: Activity
    bounty_hunter_legacy__hunter: Activity
    bounty_hunter_legacy__rogue: Activity
    clue_scrolls_all: Activity
    clue_scrolls_beginner: Activity
    clue_scrolls_easy: Activity
    clue_scrolls_medium: Activity
    clue_scrolls_hard: Activity
    clue_scrolls_elite: Activity
    clue_scrolls_master: Activity
    lms__rank: Activity
    pvp_arena__rank: Activity
    soul_wars_zeal: Activity
    rifts_closed: Activity
    abyssal_sire: Activity
    alchemical_hydra: Activity
    artio: Activity
    barrows_chests: Activity
    bryophyta: Activity
    callisto: Activity
    calvarion: Activity
    cerberus: Activity
    chambers_of_xeric: Activity
    chambers_of_xeric__challenge_mode: Activity
    chaos_elemental: Activity
    chaos_fanatic: Activity
    commander_zilyana: Activity
    corporeal_beast: Activity
    crazy_archaeologist: Activity
    dagannoth_prime: Activity
    dagannoth_rex: Activity
    dagannoth_supreme: Activity
    deranged_archaeologist: Activity
    duke_sucellus: Activity
    general_graardor: Activity
    giant_mole: Activity
    grotesque_guardians: Activity
    hespori: Activity
    kalphite_queen: Activity
    king_black_dragon: Activity
    kraken: Activity
    kreearra: Activity
    kril_tsutsaroth: Activity
    mimic: Activity
    nex: Activity
    nightmare: Activity
    phosanis_nightmare: Activity
    obor: Activity
    phantom_muspah: Activity
    sarachnis: Activity
    scorpia: Activity
    scurrius: Activity
    skotizo: Activity
    spindel: Activity
    tempoross: Activity
    the_gauntlet: Activity
    the_corrupted_gauntlet: Activity
    the_leviathan: Activity
    the_whisperer: Activity
    theatre_of_blood: Activity
    theatre_of_blood__hard_mode: Activity
    thermonuclear_smoke_devil: Activity
    tombs_of_amascut: Activity
    tombs_of_amascut__expert_mode: Activity
    tzkalzuk: Activity
    tztokjad: Activity
    vardorvis: Activity
    venenatis: Activity
    vetion: Activity
    vorkath: Activity
    wintertodt: Activity
    zalcano: Activity
    zulrah: Activity

    def __init__(self, username: str):
        player_data = PlayerFetcher.fetch_player_by_name(username)
        skills = player_data.get("skills", {})
        activities = player_data.get("activities", {})

        for skill in SKILLS:
            setattr(
                self,
                skill.lower(),
                Skill(
                    rank=skills.get(skill, {}).get("rank"),
                    level=skills.get(skill, {}).get("level"),
                    xp=skills.get(skill, {}).get("xp"),
                ),
            )
        for activity in ACTIVITIES:
            setattr(
                self,
                activity.lower()
                .replace(" ", "_")
                .replace("'", "")
                .replace("-", "")
                .replace("(", "")
                .replace(")", "")
                .replace(":", "_"),
                Activity(
                    rank=activities.get(activity, {}).get("rank"),
                    score=activities.get(activity, {}).get("score"),
                ),
            )



from mwparserfromhell.nodes.template import Template


def mw_safe_get(mw_template: Template, keys: list, ret_type=str):
    """Get value of any one key among keys from a MediaWiki template."""

    val = ret_type()
    for key in keys:
        if mw_template.has(key):
            val = ret_type(str(mw_template.get(key).value))
            break

    return val


def mw_safe_get_match(mw_template: Template, key, ret_type=str):
    """Get values of all matching keys from a MediaWiki template."""

    vals = list()
    print(mw_template.params)
    for param in mw_template.params:
        if key in param.name:
            vals.append(ret_type(param.value))

    return vals


def get_friendly_unit(n: int):
    """Get prices in units like 2.1b or 100k"""

    available_units = {1_000_000_000: "b", 1_000_000: "m", 1000: "k"}
    unit = ""
    place_value = 1000
    for value in available_units:
        if value <= n:
            unit = available_units[value]
            place_value = value
            break

    while n > place_value:
        n /= place_value

    return f"{format(n,'.1f')}{unit}"



from dataclasses import dataclass
import re
from typing import TypedDict

import mwparserfromhell as mw
import requests

from yew import WIKI_URL, USER_AGENT
from yew.utils import mw_safe_get, mw_safe_get_match


class NpcFetcher:
    @staticmethod
    def fetch_wikicode(name: str) -> str | None:

        url = f"{WIKI_URL}?action=query&prop=revisions&titles={name}&rvslots=*&rvprop=content&formatversion=2&format=json"
        try:
            response = requests.get(url, headers={"User-Agent": USER_AGENT}).json()

            wikicode: str = (
                response.get("query", {})
                .get("pages", [{}])[0]
                .get("revisions", [{}])[0]
                .get("slots", {})
                .get("main", {})
                .get("content")
            )
            wikicode = wikicode.replace("\n", "")

            return wikicode

        except Exception as e:
            raise e


class NpcParser:

    @staticmethod
    def parse_npc(wikicode: str | None) -> dict[str, str | bool | int | None]:

        npc = {}
        infobox: mw.nodes.template.Template = None

        regex = re.compile(r"[^a-zA-Z0-9 \-\,]")
        regex_keep_space = re.compile(r"[^a-zA-Z0-9\-\,]")

        templates = mw.parse(wikicode).filter_templates()

        for template in templates:
            if "Infobox NPC" in template.name:
                infobox = template
                break

        if not infobox:
            raise RuntimeError(f"NPC not found")

        npc["name"] = mw_safe_get(infobox, ["name", "name1"]).strip()

        members = mw_safe_get(infobox, ["members", "members1"])
        npc["members"] = True if "Yes" in members else False

        npc["race"] = regex.sub("", mw_safe_get(infobox, ["race", "race1"])).strip()

        npc["examine"] = regex.sub(
            "", mw_safe_get(infobox, ["examine", "examine1"])
        ).strip()

        npc["image"] = f"""https://oldschool.runescape.wiki/images/{npc["name"]}.png"""
        npc["level"] = mw_safe_get(infobox, ["level", "level1"], ret_type=int)

        npc["gender"] = regex.sub("", mw_safe_get(infobox, ["gender"])).strip()

        locations = mw_safe_get(infobox, ["location", "location1"]).split(",")
        locations = [regex.sub("", location).strip() for location in locations]
        npc["locations"] = locations

        return npc

    @staticmethod
    def parse_monster(wikicode: str | None):
        npc = {}
        infobox: mw.nodes.template.Template = None

        regex = re.compile(r"[^a-zA-Z0-9 \-\,]")
        regex_keep_space = re.compile(r"[^a-zA-Z0-9\-\,]")

        templates = mw.parse(wikicode).filter_templates()

        for template in templates:
            if "Infobox Monster" in template.name:
                infobox = template
                break

        if not infobox:
            raise RuntimeError(f"NPC not found")

        npc["name"] = mw_safe_get(infobox, ["name", "name1"]).strip()

        members = mw_safe_get(infobox, ["members", "members1"])
        npc["members"] = True if "Yes" in members else False

        npc["race"] = regex.sub("", mw_safe_get(infobox, ["race", "race1"])).strip()

        npc["examine"] = regex.sub(
            "", mw_safe_get(infobox, ["examine", "examine1"])
        ).strip()

        npc["image"] = f"""https://oldschool.runescape.wiki/images/{npc["name"]}.png"""
        npc["level"] = mw_safe_get(infobox, ["level", "level1", "combat"], ret_type=int)

        npc["gender"] = regex.sub("", mw_safe_get(infobox, ["gender"])).strip()

        npc["max_hit"] = mw_safe_get(infobox, ["max hit"], ret_type=int)

        aggresive = mw_safe_get(infobox, ["aggresive"])
        npc["aggresive"] = True if "Yes" in aggresive else False

        poisonous = mw_safe_get(infobox, ["poisonous"])
        npc["poisonous"] = True if "Yes" in poisonous else False

        npc["hitpoints"] = mw_safe_get(infobox, ["hitpoints", "hitpoints1"], ret_type=int)
        npc["attack"] = mw_safe_get(infobox, ["att", "att1"], ret_type=int)
        npc["strength"] = mw_safe_get(infobox, ["str", "str1"], ret_type=int)
        npc["defence"] = mw_safe_get(infobox, ["def", "def1"], ret_type=int)
        npc["magic"] = mw_safe_get(infobox, ["mage", "mage1"], ret_type=int)
        npc["ranged"] = mw_safe_get(infobox, ["range", "range1"], ret_type=int)

        locations = mw_safe_get(infobox, ["location", "location1"]).split(",")
        locations = [regex.sub("", location).strip() for location in locations]
        npc["locations"] = locations

        return npc


@dataclass
class Npc:
    name: str | None
    members: bool | None
    race: str
    examine: str
    image: str
    level: int
    gender: str
    max_hit: int
    aggresive: bool
    poisonous: bool
    hitpoints: int
    attack: int
    strength: int
    defence: int
    magic: int
    ranged: int
    locations: list[str]

    @classmethod
    def from_wiki(cls, name: str):
        wikicode = NpcFetcher.fetch_wikicode(name) or ""

        if "Infobox NPC" in wikicode:
            npc = NpcParser.parse_npc(wikicode)
        if "Infobox Monster" in wikicode:
            npc = NpcParser.parse_monster(wikicode)

        return cls(
            name=npc.get("name"),
            members=npc.get("members"),
            race=npc.get("race"),
            examine=npc.get("examine"),
            image=npc.get("image"),
            level=npc.get("level"),
            gender=npc.get("gender"),
            max_hit=npc.get("max_hit"),
            aggresive=npc.get("aggresive"),
            poisonous=npc.get("poisonous"),
            hitpoints=npc.get("hitpoints"),
            attack=npc.get("attack"),
            strength=npc.get("strength"),
            defence=npc.get("defence"),
            magic=npc.get("magic"),
            ranged=npc.get("ranged"),
            locations=npc.get("locations"),
        )



from dataclasses import dataclass, field
import json
from datetime import datetime

import requests

# from yew.utils.api import fetch_item_prices
from yew.utils import get_friendly_unit
from yew import USER_AGENT, ITEMS_URL, ITEMS, PRICES_URL


@dataclass
class Trend:
    """Price trends for an Item."""

    current: float
    today: float
    day30: float
    day90: float
    day180: float


class ItemFetcher:

    @staticmethod
    def fetch_item_by_id(item_id: int):
        item_data = {}

        item_data = ITEMS.get(item_id, {})

        url = f"{ITEMS_URL}/catalogue/detail.json?item={item_id}"
        response = (
            requests.get(url, headers={"User-Agent": USER_AGENT}).json().get("item", {})
        )
        item_data["icon"] = response.get("icon")
        item_data["type"] = response.get("type")
        item_data["current"] = response.get("current")
        item_data["today"] = response.get("today")
        item_data["day30"] = response.get("day30")
        item_data["day90"] = response.get("day90")
        item_data["day180"] = response.get("day180")

        return item_data

    @staticmethod
    def fetch_item_by_name(name: str):
        item_data = {}

        for _id in ITEMS:
            if ITEMS[_id].get("name").lower() in name.lower():
                item_data = ITEMS[_id]
                item_id = _id
                break

        url = f"{ITEMS_URL}/catalogue/detail.json?item={item_id}"
        response = (
            requests.get(url, headers={"User-Agent": USER_AGENT}).json().get("item", {})
        )
        item_data["icon"] = response.get("icon")
        item_data["type"] = response.get("type")
        item_data["current"] = response.get("current")
        item_data["today"] = response.get("today")
        item_data["day30"] = response.get("day30")
        item_data["day90"] = response.get("day90")
        item_data["day180"] = response.get("day180")

        return item_data


class PriceFetcher:

    @staticmethod
    def fetch_item_price(item_id: int, interval: str):
        if interval == "latest":
            url = f"{PRICES_URL}/latest?id={item_id}"
        else:
            url = f"{PRICES_URL}/timeseries?timestep={interval}&id={item_id}"

        response = (
            requests.get(url, headers={"User-Agent": USER_AGENT}).json().get("data", {})
        )
        return response


@dataclass()
class Price:
    item_id: int
    high_price: int
    low_price: int
    high_price_time: datetime
    low_price_time: datetime
    since: datetime
    high_price_volume: int | None = None
    low_price_volume: int | None = None
    interval: str = "latest"
    friendly_units: bool = False

    def __init__(self, item_id: int, interval: str):
        if not item_id:
            raise RuntimeError("Item ID is not provided")

        self.item_id = item_id
        self.interval = interval

        price_data = PriceFetcher.fetch_item_price(self.item_id, self.interval)
        price_data = price_data.get(f"{item_id}")

        self.high_price = price_data.get("high", 0)
        self.low_price = price_data.get("low", 0)
        self.high_price_time = datetime.fromtimestamp(price_data.get("highTime", 0))
        self.low_price_time = datetime.fromtimestamp(price_data.get("lowTime", 0))
        self.since = (
            self.high_price_time
            if self.high_price_time > self.low_price_time
            else self.low_price_time
        )


@dataclass
class Item:
    """The Item class."""

    id: int
    name: str
    members: bool
    examine: str
    value: int
    highalch: int
    lowalch: int
    icon: str
    type_: str
    limit: int
    trend: Trend
    price: Price | None = field(init=False, default=None)

    @classmethod
    def from_id(cls, id: int):
        """Fetch an item using its ID."""

        if not isinstance(id, int):
            raise TypeError(f"Expected id to be of type int, got {type(id)}")

        item_data = ItemFetcher.fetch_item_by_id(id)

        return cls(
            id=id,
            name=item_data.get("name"),
            members=True if item_data.get("members") == "true" else False,
            examine=item_data.get("examine"),
            value=item_data.get("value"),
            highalch=item_data.get("highalch"),
            lowalch=item_data.get("lowalch"),
            icon=item_data.get("icon"),
            type_=item_data.get("type"),
            limit=item_data.get("limit"),
            trend=Trend(
                current=item_data.get("current"),
                today=item_data.get("today"),
                day30=item_data.get("day30"),
                day90=item_data.get("day90"),
                day180=item_data.get("day180"),
            ),
        )

    @classmethod
    def from_name(cls, name: str):
        """Fetch an item using its name."""

        if not isinstance(name, str):
            raise TypeError(f"Expected name to be of type str, got {type(name)}")

        item_data = ItemFetcher.fetch_item_by_name(name)

        return cls(
            id=item_data.get("item_id"),
            name=item_data.get("name"),
            members=item_data.get("members"),
            examine=item_data.get("examine"),
            value=item_data.get("value"),
            highalch=item_data.get("highalch"),
            lowalch=item_data.get("lowalch"),
            icon=item_data.get("icon"),
            type_=item_data.get("type"),
            limit=item_data.get("limit"),
            trend=Trend(
                current=item_data.get("current"),
                today=item_data.get("today"),
                day30=item_data.get("day30"),
                day90=item_data.get("day90"),
                day180=item_data.get("day180"),
            ),
        )

    def prices(self, interval: str):
        self.price = Price(self.id, interval)
        return self.price
